// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013-2014, Gimpy Software.

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace LayersBlog
{
    /// <summary>
    /// Interface wrapping a GPU-based XNA graphics effect.
    /// </summary>
    /// <remarks>
    /// Creating an interface allows us to derive classes which include effect-specific setup.
    /// For example, a Gaussian blur implementation could set up lookup tables required
    /// by the .fx code.
    /// </remarks>
    public abstract class GpuLayerEffect : ILayerEffect
    {
        protected Effect Effect { get; set; }

        public virtual void SetParameters(Texture2D source)
        {
            var parameter = Effect.Parameters["ImageWidth"];
            if (parameter != null)
                parameter.SetValue(source.Width);
            parameter = Effect.Parameters["ImageHeight"];
            if (parameter != null)
                parameter.SetValue(source.Height);
        }

        public void Apply(SpriteBatch spriteBatch, Texture2D source, Texture2D destination)
        {
            // Write to the SpriteBatch destination using the Effect.
            spriteBatch.Begin(SpriteSortMode.Immediate, null, null, null, null, Effect);
            spriteBatch.Draw(source, Vector2.Zero, Color.White);
            spriteBatch.End();
        }
    }

    public class TextureBuffer
    {
        protected int m_bufferWidth;
        private int m_bufferHeight;
        private Color[] m_colors;

        public void SetSize(int width, int height)
        {
            if (m_colors == null || m_bufferWidth < width || m_bufferHeight < height)
            {
                m_colors = new Color[width * height];
                m_bufferWidth = width;
                m_bufferHeight = height;
            }
        }

        public Color[] Buffer
        {
            get { return m_colors; }
        }

        public int GetIndexFromCoord(int x, int y)
        {
#if DEBUG
            if (x < 0 || x >= m_bufferWidth)
                throw new ArgumentOutOfRangeException("x");
            if (y < 0 || y >= m_bufferHeight)
                throw new ArgumentOutOfRangeException("y");
#endif

            return y * m_bufferWidth + x;
        }
    }

    /// <summary>
    /// A 'simple' effect, directly wrapping an XNA Effect object.
    /// </summary>
    public class SimpleEffect : GpuLayerEffect
    {
        public SimpleEffect(Effect effect)
        {
            Effect = effect;
        }
    }
}