// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013-2014, Gimpy Software.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace LayersBlog
{
    /// <summary>
    /// The base class for a single 'layer'.
    /// </summary>
    /// <remarks>
    /// This provides functionality for defining the basic appearance of a layer.
    /// </remarks>
    public abstract class Layer : IDisposable
    {
        protected readonly GraphicsDevice m_graphicsDevice;
        private readonly int m_width;
        private readonly int m_height;
        private List<ILayerEffect> m_effects;

        private RenderTarget2D m_effectTarget1;
        private RenderTarget2D m_effectTarget2;
        private int m_targetIndex;

        /// <summary>
        /// Gets the parent Layer object, if defined.
        /// </summary>
        public Layer Parent { get; protected internal set; }

        /// <summary>
        /// Gets or sets the opacity for this layer.
        /// </summary>
        /// <remarks>
        /// 0.0f is fully transparent, 1.0f is fully opaque.
        /// </remarks>
        public float Opacity { get; set; }

        /// <summary>
        /// The pixel origin of the layer.
        /// </summary>
        /// <remarks>
        /// I.e. Increasing X will move the image to the left.
        /// </remarks>
        public Vector2 Origin { get; set; }

        /// <summary>
        /// Gets or sets whether this layer should be sent draw and update calls.
        /// </summary>
        /// <remarks>
        /// It is true by default, meaning the layer will be processed as part of a Layers group.
        /// </remarks>
        public bool IsEnabled { get; set; }

        public List<ILayerEffect> Effects
        {
            get { return m_effects ?? (m_effects = new List<ILayerEffect>()); }
        }

        internal RenderTarget2D ActiveTarget
        {
            get
            {
                return m_targetIndex == 0 ? EffectTarget1 : EffectTarget2;
            }
        }

        private RenderTarget2D AltTarget
        {
            get
            {
                return m_targetIndex == 1 ? EffectTarget1 : EffectTarget2;
            }
        }

        private RenderTarget2D EffectTarget1
        {
            get { return m_effectTarget1 ?? (m_effectTarget1 = new RenderTarget2D(m_graphicsDevice, m_width, m_height)); }
        }

        private RenderTarget2D EffectTarget2
        {
            get { return m_effectTarget2 ?? (m_effectTarget2 = new RenderTarget2D(m_graphicsDevice, m_width, m_height)); }
        }

        private void SwapTargets()
        {
            m_targetIndex = (m_targetIndex + 1) % 2;
        }

        protected Layer(GraphicsDevice graphicsDevice, int width, int height)
        {
            m_graphicsDevice = graphicsDevice;
            m_width = width;
            m_height = height;

            IsEnabled = true;
            Opacity = 1.0f;

            BackgroundFillColor = Color.Transparent;
        }

        /// <summary>
        /// External code will call this to draw the layer.
        /// </summary>
        public void Draw(SpriteBatch spriteBatch)
        {
            Render(spriteBatch);

            spriteBatch.Begin();
            spriteBatch.Draw(ActiveTarget, Vector2.Zero, Color.White);
            spriteBatch.End();
        }

        /// <summary>
        /// Render the layer (including effects) onto the ActiveTarget.
        /// </summary>
        protected internal virtual void Render(SpriteBatch spriteBatch)
        {
            // Draw the 'plain' layer to a temporary render target.
            m_graphicsDevice.SetRenderTarget(ActiveTarget);
            m_graphicsDevice.Clear(BackgroundFillColor);
            DrawImpl(spriteBatch);
            m_graphicsDevice.SetRenderTarget(null);

            // Apply each effect in turn.
            ApplyEffects(spriteBatch);
        }

        public Color BackgroundFillColor { get; set; }

        protected void ApplyEffects(SpriteBatch spriteBatch)
        {
            foreach (var effect in Effects)
            {
                SwapTargets();
                DrawWithEffect(spriteBatch, AltTarget, ActiveTarget, effect);
            }
        }

        private void DrawWithEffect(SpriteBatch spriteBatch, Texture2D source, RenderTarget2D destination, ILayerEffect layerEffect)
        {
            m_graphicsDevice.SetRenderTarget(destination);
            m_graphicsDevice.Clear(Color.Transparent);

            var gpuLayerEffect = layerEffect as GpuLayerEffect;
            if (gpuLayerEffect != null)
                gpuLayerEffect.SetParameters(source);

            layerEffect.Apply(spriteBatch, source, destination);

            m_graphicsDevice.SetRenderTarget(null);
        }

        protected static void Save(Texture2D destination, string file)
        {
            using (var stream = new FileStream(file, FileMode.OpenOrCreate))
                destination.SaveAsPng(stream, destination.Width, destination.Height);
        }

        /// <summary>
        /// Implementations should draw their layer content here.
        /// </summary>
        protected abstract void DrawImpl(SpriteBatch spriteBatch);

        /// <summary>
        /// Implementations can update their state here.
        /// </summary>
        /// <remarks>
        /// The default implementation is a no-op.
        /// </remarks>
        public virtual void Update(GameTime gameTime)
        {
        }

        /// <summary>
        /// Utility function for creating a white color with a specific amount of opacity.
        /// </summary>
        protected static Color CreateWhite(float opacity)
        {
            return new Color(1.0f, 1.0f, 1.0f, opacity);
        }

        public virtual void Dispose()
        {
            if (m_effectTarget1 != null)
            {
                m_effectTarget1.Dispose();
                m_effectTarget1 = null;
            }

            if (m_effectTarget2 != null)
            {
                m_effectTarget2.Dispose();
                m_effectTarget2 = null;
            }
        }
    }

    /// <summary>
    /// A collection of Layer objects.
    /// </summary>
    /// <remarks>
    /// As we inherit from Layer a Layers objects can be nested.
    /// In the simplest case you would create one of these objects, populate it with
    /// layers, and call Draw(...) in your render code.
    /// </remarks>
    public class Layers : Layer
    {
        private readonly List<Layer> m_layers = new List<Layer>();

        public Layers(GraphicsDevice graphicsDevice, int width, int height) : base(graphicsDevice, width, height)
        {
        }

        protected internal override void Render(SpriteBatch spriteBatch)
        {
            // Render each layer.
            foreach (var layer in m_layers.Where(layer => layer.IsEnabled))
                layer.Render(spriteBatch);

            // Combine layers onto ActiveTarget.
            m_graphicsDevice.SetRenderTarget(ActiveTarget);
            DrawImpl(spriteBatch);
            m_graphicsDevice.SetRenderTarget(null);

            // Apply each effect in turn.
            ApplyEffects(spriteBatch);
        }

        protected override void DrawImpl(SpriteBatch spriteBatch)
        {
            spriteBatch.Begin();
            foreach (var layer in m_layers.Where(layer => layer.IsEnabled))
                spriteBatch.Draw(layer.ActiveTarget, Vector2.Zero, Color.White);
            spriteBatch.End();
        }

        /// <summary>
        /// Updates each 'enabled' layer.
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Update(GameTime gameTime)
        {
            foreach (var layer in m_layers.Where(layer => layer.IsEnabled))
                layer.Update(gameTime);

            base.Update(gameTime);
        }

        /// <summary>
        /// Add a Layer to the collection.
        /// </summary>
        /// <remarks>
        /// Layers will be added 'on top of' any previous layers.
        /// </remarks>
        public void Add(Layer layer)
        {
            layer.Parent = this;
            m_layers.Add(layer);
        }

        /// <summary>
        /// Remove the specified layer from the collection.
        /// </summary>
        public virtual void Remove(Layer layer)
        {
            m_layers.Remove(layer);
        }

        /// <summary>
        /// Remove all layers from the collection.
        /// </summary>
        public virtual void Clear()
        {
            m_layers.Clear();
        }
    }

    /// <summary>
    /// A Layer implementation which displays a Texture2D object to the viewport origin.
    /// </summary>
    public class TextureLayer : Layer
    {
        private readonly Texture2D m_texture;

        public TextureLayer(Texture2D texture) : base(texture.GraphicsDevice, texture.Width, texture.Height)
        {
            m_texture = texture;
        }

        protected override void DrawImpl(SpriteBatch spriteBatch)
        {
            spriteBatch.Begin();
            spriteBatch.Draw(m_texture, Origin, CreateWhite(Opacity));
            spriteBatch.End();
        }
    }

    /// <summary>
    /// A Layer implementation which displays on-screen FPS and garbage collection information.
    /// </summary>
    public class StatsLayer : Layer
    {
        private readonly SpriteFont m_font;
        private Stopwatch m_stopwatch;
        private int m_frameCount;

        /// <summary>
        /// Constructor, specifying the SpriteFont to use to display the statistics.
        /// </summary>
        public StatsLayer(GraphicsDevice graphicsDevice, int width, int height, SpriteFont font) : base(graphicsDevice, width, height)
        {
            m_font = font;
        }

        public override void Update(GameTime gameTime)
        {
            base.Update(gameTime);

            if (m_stopwatch == null)
                m_stopwatch = Stopwatch.StartNew();

            if (m_stopwatch.ElapsedMilliseconds >= 500)
            {
                FPS = (int)(m_frameCount / (m_stopwatch.ElapsedMilliseconds / 1000.0f));
                m_stopwatch.Restart();
                m_frameCount = 0;
            }
        }

        private int FPS { get; set; }

        protected override void DrawImpl(SpriteBatch spriteBatch)
        {
            m_frameCount++;
            var s = string.Format("{0} FPS\n{1} GC Gen 0\n{2} GC Gen 1\n{3} GC Gen 2", FPS, GC.CollectionCount(0), GC.CollectionCount(1), GC.CollectionCount(2));

            spriteBatch.Begin();
            spriteBatch.DrawString(m_font, s, Vector2.Zero, Color.Yellow);
            spriteBatch.End();
        }
    }

    /// <summary>
    /// A disposable class which allows you to display a video as a layer.
    /// </summary>
    public class VideoLayer : Layer
    {
        private readonly Video m_video;
        private readonly bool m_autoPlay;
        private VideoPlayer m_player;
        private bool m_isFirstUpdate = true;

        /// <summary>
        /// Gets or sets a value indicating whether to automatically loop the video.
        /// </summary>
        public bool AutoRepeat
        {
            get { return m_player.IsLooped; }
            set { m_player.IsLooped = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="video">The Video object to play.</param>
        /// <param name="autoPlay">Whether to start as video playing as soon as Update(...) is called.</param>
        public VideoLayer(GraphicsDevice graphicsDevice, Video video, bool autoPlay) : base(graphicsDevice, video.Width, video.Height)
        {
            m_video = video;
            m_autoPlay = autoPlay;
            m_player = new VideoPlayer();
            AutoRepeat = true;
        }

        public void Play()
        {
            m_player.Stop();
            m_player.Play(m_video);
        }

        public void Stop()
        {
            m_player.Stop();
        }

        public override void Update(GameTime gameTime)
        {
            if (m_isFirstUpdate)
            {
                if (m_autoPlay)
                    Play();
                m_isFirstUpdate = false;
            }

            base.Update(gameTime);
        }

        protected override void DrawImpl(SpriteBatch spriteBatch)
        {
            spriteBatch.GraphicsDevice.Clear(Color.Black);

            if (m_player.State != MediaState.Stopped)
            {
                spriteBatch.Begin();
                spriteBatch.Draw(m_player.GetTexture(), Origin, CreateWhite(Opacity));
                spriteBatch.End();
            }
        }

        public override void Dispose()
        {
            if (m_player != null)
            {
                m_player.Dispose();
                m_player = null;
            }

            base.Dispose();
        }
    }

    /// <summary>
    /// A collection of layers, each of which has its movement scaled.
    /// </summary>
    /// <remarks>
    /// This allows layers representing distant scene objects to move slower than those 'nearer' to the viewer.
    /// </remarks>
    public class ParallaxLayer : Layers
    {
        private class LayerInfo
        {
            public Layer Layer { get; private set; }
            public float MovementFactor { get; private set; }

            public LayerInfo(Layer layer, float movementFactor)
            {
                Layer = layer;
                MovementFactor = movementFactor;
            }
        }

        private readonly List<LayerInfo> m_layerInfos = new List<LayerInfo>();

        public ParallaxLayer(GraphicsDevice graphicsDevice, int width, int height) : base(graphicsDevice, width, height)
        {
        }

        /// <summary>
        /// Add a layer to the scene.
        /// </summary>
        /// <param name="layer">The layer to add.</param>
        /// <param name="movementFactor">
        /// How much to move the layer in relation to the others.
        /// (E.g. 0.5f will move the layer half as much as a layer with value 1.0f.)
        /// </param>
        public void Add(Layer layer, float movementFactor)
        {
            Add(layer);
            m_layerInfos.Add(new LayerInfo(layer, movementFactor));
        }

        public override void Remove(Layer layer)
        {
            base.Remove(layer);
            m_layerInfos.Remove(m_layerInfos.First(o => o.Layer == layer));
        }

        public override void Clear()
        {
            base.Clear();
            m_layerInfos.Clear();
        }

        public override void Update(GameTime gameTime)
        {
            foreach (var layerInfo in m_layerInfos)
                layerInfo.Layer.Origin = Origin * layerInfo.MovementFactor;

            base.Update(gameTime);
        }
    }

    /// <summary>
    /// A layer which displays text from a Logger object.
    /// </summary>
    /// <remarks>
    /// To toggle the visibility of the log the SCROLL LOCK key.
    /// To move to the start/end of the log use HOME/END.
    /// To scroll through the log use PAGE UP/PAGE DOWN.
    /// </remarks>
    public class LoggerLayer : Layer
    {
        private readonly SpriteFont m_font;
        private readonly Logger m_logger;
        private readonly int m_screenLines;
        private double m_firstLine;
        private bool m_isVisible;
        private KeyboardState m_previousState;

        public LoggerLayer(GraphicsDevice graphicsDevice, int width, int height, SpriteFont font, Logger logger) : base(graphicsDevice, width, height)
        {
            m_font = font;
            m_logger = logger;

            m_screenLines = height / font.LineSpacing;
            m_previousState = Keyboard.GetState();
        }

        public override void Update(GameTime gameTime)
        {
            const int ScrollSpeed = 20;

            base.Update(gameTime);

            var state = Keyboard.GetState();
            try
            {
                if (state.IsKeyDown(Keys.Scroll) && !m_previousState.IsKeyDown(Keys.Scroll))
                    m_isVisible = !m_isVisible; 
            
                if (!m_isVisible)
                    return;

                if (state.IsKeyDown(Keys.PageDown))
                    m_firstLine += gameTime.ElapsedGameTime.TotalSeconds * ScrollSpeed;
                else if (state.IsKeyDown(Keys.PageUp))
                    m_firstLine = Math.Max(0, m_firstLine - gameTime.ElapsedGameTime.TotalSeconds * ScrollSpeed);
                else if (state.IsKeyDown(Keys.Home))
                    m_firstLine = 0;
                else if (state.IsKeyDown(Keys.End))
                    m_firstLine = Math.Max(0, m_logger.Messages.Count - m_screenLines);
            }
            finally
            {
                m_previousState = state;
            }
        }

        protected override void DrawImpl(SpriteBatch spriteBatch)
        {
            if (!m_isVisible)
                return;

            m_graphicsDevice.Clear(new Color(0, 0, 0, 0.6f));

            var position = new Vector2(0, m_font.LineSpacing);
            var topLineIndex = (int)m_firstLine;
            var screenLine = 0;

            spriteBatch.Begin();
            foreach (var message in m_logger.Messages.Skip(topLineIndex).Take(m_screenLines))
                spriteBatch.DrawString(m_font, message, position * screenLine++, Color.White);
            spriteBatch.End();
        }
    }
}
