// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013-2014, Gimpy Software.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace LayersBlog
{
    /// <summary>
    /// A class to allow application messages to be logged.
    /// </summary>
    /// <remarks>
    /// All messages will be stored (which can be cleared by calling Clear()), and each will also
    /// be output to the debugger's output window.
    /// </remarks>
    public class Logger
    {
        private List<string> m_messages;

        public List<string> Messages
        {
            get { return m_messages ?? (m_messages = new List<string>()); }
        }

        /// <summary>
        /// Add a new message to the log.
        /// </summary>
        /// <remarks>
        /// Message strings containing carriage returns ('\n') will be spit into multiple message lines.
        /// </remarks>
        /// <param name="message"></param>
        public void Log(string message)
        {
            foreach (var line in message.Split(new[] { '\n' }, StringSplitOptions.None))
            {
                Messages.Add(line);
                Debug.WriteLine(line);
            }
        }

        /// <summary>
        /// Clear all messages.
        /// </summary>
        public void Clear()
        {
            Messages.Clear();
        }

        /// <summary>
        /// Logs information about the current environment.
        /// </summary>
        public void LogEnvironment()
        {
            var now = DateTime.Now;
            Log(now.ToLongDateString() + ", " + now.ToLongTimeString());
            LogUnderline();
        }

        /// <summary>
        /// Automatically adds an underline (using '-' characters) to the previous log message.
        /// </summary>
        /// <remarks>
        /// This is a no-op if there are no previous messages.
        /// </remarks>
        public void LogUnderline()
        {
            if (Messages.Count == 0)
                return;
            Log(new string('-', Messages.Last().Length));
        }
    }
}