// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013-2014, Gimpy Software.

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace LayersBlog
{
    /// <summary>
    /// An effect which overlays a chequered pattern over transparent pixels.
    /// </summary>
    public class ShowTransparencyEffect : SimpleEffect
    {
        public ShowTransparencyEffect(ContentManager content)
            : base(content.Load<Effect>("FX/showTransparency"))
        {
            GridSize = 10;
        }

        /// <summary>
        /// Gets or sets the size of the pattern, in pixels.
        /// </summary>
        public int GridSize { get; set; }

        public override void SetParameters(Texture2D source)
        {
            base.SetParameters(source);

            Effect.Parameters["SegmentsPerXAxis"].SetValue((int)(1.0f * source.Width / GridSize));
            Effect.Parameters["SegmentsPerYAxis"].SetValue((int)(1.0f * source.Height / GridSize));
        }
    }

    /// <summary>
    /// Appears to decrease the resolution of an image by a variable amount.
    /// </summary>
    public class PixelateEffect : SimpleEffect
    {
        public PixelateEffect(ContentManager content)
            : base(content.Load<Effect>("FX/pixelate"))
        {
            GridSize = 20;
        }

        /// <summary>
        /// Gets or sets the size of the 'apparent' pixels, in pixels.
        /// </summary>
        public int GridSize { get; set; }

        public override void SetParameters(Texture2D source)
        {
            base.SetParameters(source);

            Effect.Parameters["SegmentsPerXAxis"].SetValue((int)(1.0f * source.Width / GridSize));
            Effect.Parameters["SegmentsPerYAxis"].SetValue((int)(1.0f * source.Height / GridSize));
        }
    }

    /// <summary>
    /// Adds 'retro' style scanlines to an image.
    /// </summary>
    public class ScanlineEffect : SimpleEffect
    {
        public ScanlineEffect(ContentManager content)
            : base(content.Load<Effect>("FX/scanlines"))
        {
        }
    }

    /// <summary>
    /// Converts the image colors to a sepia tone.
    /// </summary>
    public class SepiaEffect : SimpleEffect
    {
        public SepiaEffect(ContentManager content)
            : base(content.Load<Effect>("FX/sepia"))
        {
        }
    }

    /// <summary>
    /// Converts the image colors to grayscale.
    /// </summary>
    public class GrayscaleEffect : SimpleEffect
    {
        public GrayscaleEffect(ContentManager content)
            : base(content.Load<Effect>("FX/grayscale"))
        {
        }
    }

    /// <summary>
    /// Inverts the image colors.
    /// </summary>
    public class NegativeEffect : SimpleEffect
    {
        public NegativeEffect(ContentManager content)
            : base(content.Load<Effect>("FX/negative"))
        {
        }
    }

    /// <summary>
    /// Shows opaque pixels in black, all others transparent.
    /// </summary>
    public class OpacityMaskEffect : SimpleEffect
    {
        public OpacityMaskEffect(ContentManager content)
            : base(content.Load<Effect>("FX/opacityMask"))
        {
        }
    }

    /// <summary>
    /// Remaps all image colors to a specified target color.
    /// </summary>
    public class SaturateEffect : SimpleEffect
    {
        public SaturateEffect(ContentManager content)
            : base(content.Load<Effect>("FX/saturate"))
        {
            Target = Color.Red;
            Saturation = 0.5f;
        }

        /// <summary>
        /// The amount to change the existing colors.
        /// </summary>
        /// <remarks>
        /// 0.0f makes no change in the image.
        /// 1.0f makes all pixels the Target color.
        /// </remarks>
        public float Saturation
        {
            get { return Effect.Parameters["Saturation"].GetValueSingle(); }
            set { Effect.Parameters["Saturation"].SetValue(value); }
        }

        /// <summary>
        /// The color to map towards.
        /// </summary>
        public Color Target
        {
            get
            {
                var rgb = Effect.Parameters["TargetColor"].GetValueVector4();
                return new Color(rgb);
            }

            set
            {
                Effect.Parameters["TargetColor"].SetValue(value.ToVector4());
            }
        }
    }

    /// <summary>
    /// Reduce the effective color palette by specifying how many bits per r/g/b color channel to use.
    /// </summary>
    public class PalettizerEffect : SimpleEffect
    {
        private int m_bitsPerChannel;

        public PalettizerEffect(ContentManager content)
            : base(content.Load<Effect>("FX/palettizer"))
        {
            BitsPerChannel = 2;
        }

        public int BitsPerChannel
        {
            get
            {
                return m_bitsPerChannel;
            }

            set
            {
                m_bitsPerChannel = value;
                DiscreteValues = (int)Math.Pow(2, value);
            }
        }

        private int DiscreteValues
        {
            get { return Effect.Parameters["DiscreteValues"].GetValueInt32(); }
            set { Effect.Parameters["DiscreteValues"].SetValue(value); }
        }
    }
}